'''Timelines help organise data longitudinally'''

from pattools.pacs import Series, Patient
from pattools.resources import Atlas
from pattools.image import histogram_match, normalize_by_whitematter, estimate_window
import nibabel as nib
import numpy as np
import json
import os
import shutil
from clint.textui import progress
from tempfile import TemporaryDirectory
from datetime import date, timedelta, datetime

#local imports
import pattools._timelinefilter

class FileMetadata:
    '''Metadata for a stored image file'''
    file = None
    processed_file = None
    filter_name = None
    study_date = None
    study_uid = None
    series_uid = None
    series_description = None

    def __init__(self, file=None, processed_file=None, filter_name=None, study_date=None, study_uid=None, series_uid=None, series_description=None):
        self.file = file
        self.processed_file = processed_file
        self.filter_name = filter_name
        self.study_date = study_date
        self.study_uid = study_uid
        self.series_uid = series_uid
        self.series_description = series_description

    def __str__(self):
        return (
            'file              : ' + str(self.study_uid) +os.linesep+
            'processed_file    : ' + str(self.processed_file) +os.linesep+
            'filter_name       : ' + str(self.filter_name) +os.linesep+
            'study_date       : ' + str(self.study_date) +os.linesep+
            'study_uid         : ' + str(self.study_uid) +os.linesep+
            'series_uid        : ' + str(self.series_uid) +os.linesep+
            'series_description: ' + str(self.series_description))

    @staticmethod
    def from_string(string):
        fm = FileMetadata
        for line in string.splitlines():
            if line.startswith('file'):
                fm.file = line.split(':')[1].strip()
            elif line.startswith('processed_file'):
                fm.processed_file = line.split(':')[1].strip()
            elif line.startswith('filter_name'):
                fm.filter_name = line.split(':')[1].strip()
            elif line.startswith('study_uid'):
                fm.study_uid = line.split(':')[1].strip()
            elif line.startswith('series_uid'):
                fm.series_uid = line.split(':')[1].strip()
            elif line.startswith('study_date'):
                fm.study_date = line.split(':')[1].strip()
            if line.startswith('series_description'):
                fm.series_description = line.split(':')[1].strip()
        return fm

class Timeline:
    '''The timeline contains filtered lists of scans over a single patient's history'''
    patient_id = None #ID of the Patient in PACS
    patient_name = None
    patient_dob = None
    path = None #Path to the root timeline folder
    start_date = None #First date covered by timeline
    end_date = None #Last date covered by timeline
    brain_mask = None #Brain mask which will be used\
    whitematter_mask = None #Whitematter mask
    registration_reference = None #Reference scan for registration
    is_processed = False #Is the pre-processing up to date?
    is_rendered = False
    datamap = None #In-memory map of the data structure
    #^^ for now we'll try to use the file system to guide us
    manual_studies = None #This is a collection of studies which have been manually set to override the autmated matching.
    filters = None #Types of scans to include (defaut FLAIR and MPRAGE)
    normed_window_map = None

    def __init__(self, path, patient_id=None):
        self.path = os.path.abspath(path)
        self.patient_id = patient_id
        # If we don't have a patient ID we're assuming patient id as the folder name
        if patient_id == None:
            self.patient_id = os.path.basename(os.path.normpath(path))

        # Try to load from path...
        self.load()

        # If that doesn't work, try to create from PACS
        if not os.path.exists(path): os.makedirs(path, mode=0o777)

        # Hot tip: If we declare these at a class scope then __dict___ (and vars()) won't contain the entries.
        # Python will assume that the haven't changed because the reference hasn't changed.
        self.datamap = {}
        self.manual_studies = {}
        self.filters = pattools._timelinefilter.default_filters()
        self.normed_window_map = {}
        self.save()

    def clean(self):
        '''Removes any folders where the number of metadata files and image files don't match'''
        self.datamap = {}
        print('self.path',self.path)
        print(os.path.exists(self.path))
        if not os.path.exists(self.path): return

        to_clean = []
        for root, folders, files in os.walk(self.path):
            #Clean the directory if the stucture isn't well formed
            if (len(folders) == 0
                and len([f for f in files if f.endswith('.nii.gz') or f.endswith('.nii')])
                    < len([f for f in files if f.endswith('.metadata') and f != 'timeline.metadata'])):
                to_clean.append(os.path.join(self.path, root))
            # Or if there are no matching studies
            if len(folders) == 0 and len(files) == 0:
                to_clean.append(os.path.join(self.path, root))

        print('cleaning:', to_clean)
        for root in to_clean:
            shutil.rmtree(root)

    def update_from_pacs(self, scp_settings):
        '''Populate the Timeline from PACS'''
        print('Updating from PACS...')

        if scp_settings == None:
            print('No SCP Settings found.')
            return

        patient = Patient(self.patient_id, scp_settings)
        self.patient_name = patient.name
        self.patient_dob = patient.dob
        # Do we have new dates?

        # Sometimes there will be studies which showed up and created metadata,
        # but don't have an image. Then they dissapear and ruin everything.
        # TODO: Find out why.
        # For now we'll just clean directories where the metadata count != the image count.
        self.clean()
        try:
            print('finding studies...')
            for study in patient.find_studies():
                study_path = os.path.join(self.path, study.study_date)
                print('study path:', study_path)
                try:
                    os.mkdir(study_path)
                except:
                    pass

                # Create a new in-memory data map
                self.datamap[study.study_date] = []
                print('study date:', study.study_date)
                # Get filtered series
                for filter in self.filters:
                    series = None
                    # check for a manually selected series for the study / filter combo
                    if (filter.name, study.study_date) in self.manual_studies:
                        series = self.manual_studies[filter.name, study.study_date]
                    else:
                        series = filter.filter(study)

                    if series != None:
                        print('series:', series.description)
                        data = FileMetadata(file=filter.name + ".nii.gz")
                        new_series = True
                        metadatafile = os.path.join(study_path, data.file + '.metadata')
                        # Update existing metadata
                        if os.path.exists(metadatafile):
                            print('metadata found')
                            with open(metadatafile, 'r') as f:
                                try:
                                    data.__dict__ = json.loads(f.read())
                                except:
                                    raise Exception('Failed to read ' + metadatafile)

                            # If the series has changed, we'll delete the old one.
                            if data.series_uid != series.series_uid:
                                if os.path.exists(os.path.join(study_path, data.file)):
                                    os.remove(os.path.join(study_path, data.file))
                                if os.path.exists(os.path.join(study_path, data.processed_file)):
                                    os.remove(os.path.join(study_path, data.processed_file))
                            else:
                                # There has been no change and the file exists.
                                if os.path.exists(os.path.join(study_path, data.file)):
                                    print('nifti found and has not changed')
                                    new_series = False

                            self.datamap[study.study_date].append(data)
                        # If we have a new (or replaced) series, update everything and get the data
                        if new_series:
                            data = FileMetadata(
                                file=filter.name + ".nii.gz",
                                processed_file=filter.name + ".processed.nii.gz",
                                study_date=study.study_date,
                                filter_name=filter.name,
                                study_uid=series.study_uid,
                                series_uid=series.series_uid,
                                series_description=series.description)
                            self.datamap[study.study_date].append(data)
                            # Write metadata
                            with open(metadatafile, 'w') as f:
                                f.write(json.dumps(vars(data)))
                                f.flush()
                            print('downloading:',os.path.join(study_path,data.file))
                            series.save_nifti(os.path.join(study_path,data.file))
                            if os.path.exists(os.path.join(study_path, data.file)):
                                print('success')
                            else:
                                print('failed')

                        # Try to re-download original file if we don't have it
                        if not os.path.exists(os.path.join(study_path, data.file)):
                            print("Can't find the series, let's try again...")
                            series.save_nifti(os.path.join(study_path,data.file))

            self.is_processed = False
            self.save()
        except Exception as e:
            print('Error occurred while updating from PACS', e)

    def save(self):
        '''Save to disk'''
        content = json.dumps(vars(self), default=lambda o: o.__dict__, sort_keys=True, indent=4)
        with open(os.path.join(self.path, 'timeline.metadata'), 'w') as f:
            f.write(content)
        self._save_datamap()

    def load(self):
        '''Load from disk'''
        if not os.path.exists(os.path.join(self.path, 'timeline.metadata')):
            print('No metadata file exists in ', os.path.join(self.path, 'timeline.metadata'))
            return

        with open(os.path.join(self.path, 'timeline.metadata'), 'r') as f:
            content = f.read()
            path = self.path
            self.__dict__ = json.loads(content)
            self.path = path # The saved path may be different but this is where we loaded it.
            #These filters will load as dicts but we want them to not be. So we need to parse 'em.
            parsed_filters = []
            for filter_dict in self.__dict__['filters']:
                f = pattools._timelinefilter.Filter(filter_dict['name'])
                for key in filter_dict:
                    setattr(f,key, filter_dict[key])
                parsed_filters.append(f)
            self.filters = parsed_filters
        self._load_datamap()

    def _save_datamap(self):
        '''Save just the datamap metadata'''
        for studydate in self.datamap:
            for filemeta in self.datamap[studydate]:
                # Create study directory if it's not there
                studypath = os.path.join(self.path, studydate)
                if not os.path.exists(studypath): os.makedirs(studypath, mode=0o777)
                # Save metadata
                with open(os.path.join(studypath, filemeta.file + '.metadata'), 'w') as f:
                    f.write(json.dumps(vars(filemeta)))

    def _load_datamap(self):
        '''Load just the datamap metadata'''
        for studydate in next(os.walk(self.path))[1]:
            self.datamap[studydate] = []
            files = next(os.walk(os.path.join(self.path, studydate)))[2]
            files = [f for f in files if f.endswith('.metadata')]
            for f in files:
                with open(os.path.join(self.path, studydate, f), 'r') as f:
                    filemeta = FileMetadata()
                    try:
                        filemeta.__dict__ = json.loads(f.read())
                        self.datamap[studydate].append(filemeta)
                    except:
                        raise Exception('Failed to read ' + os.path.join(self.path, studydate, f))


    def setup_registration_reference(self):
        '''Select a registration reference from the image data'''
        from pattools import ants
        print('Setting up registration reference...')
        # Check that we don't already have one
        if (self.registration_reference != None
            and os.path.exists(os.path.join(self.path,self.registration_reference))
            and self.brain_mask != None
            and os.path.exists(os.path.join(self.path,self.brain_mask))):
            return
        # Probably better ways to pick a candidate but we're going with "large but not too large"
        candidates = []
        for dir, _, files in os.walk(self.path):
            for f in files:
                if (f.endswith('.nii') or f.endswith('.nii.gz')) and f+'.metadata' in files:
                    candidates.append(os.path.join(dir, f))
        # candidates are sorted by the minimum dimension size
        candidates.sort(key=lambda c : min(nib.load(c).shape)) # Sort by minimum slices
        #candidates.sort(key=lambda c : os.stat(c).st_size) # Sort by file size
        if len(candidates) == 0: return
        # return the candidate with the largest minimum dimension size
        candidate = candidates[int(len(candidates) * 3 / 4)]
        print('candidate:', candidate)

        with TemporaryDirectory() as tmp_dir:
            #Open atlas
            atlas_path = os.path.join(self.path, '../atlas/mni')
            if not os.path.exists(atlas_path): os.makedirs(atlas_path, mode=0o777)
            atlas = Atlas.MNI.load(atlas_path)

            # save atlas to tmp_dir and mask to timeline
            t2_path = os.path.join(tmp_dir, 't2.nii.gz')
            mask_path = os.path.join(tmp_dir, 'mask.nii.gz')
            whitematter_mask_path = os.path.join(tmp_dir, 'whitematter_mask.nii.gz')
            nib.save(atlas.t2, t2_path)
            nib.save(atlas.mask, mask_path)
            nib.save(atlas.whitematter_mask, whitematter_mask_path)

            # Bias correction and registration
            print('        N4 Bias correction for reference image...')
            n4_path = os.path.join(self.path, 'registration_reference.nii.gz')
            out_path = os.path.join(tmp_dir, 'warped.nii.gz')
            ants.n4_bias_correct(candidate, n4_path).wait()

            print('        Registering brain mask to reference image...')
            # Register mask to reference scan
            ants.affine_registration(t2_path, n4_path, out_path).wait()
            # These will be the output of the registration
            affine_mat = out_path + '_0GenericAffine.mat'
            #inverse_warp = out_path + '_1InverseWarp.nii.gz'
            #warp = out_path + '_1Warp.nii.gz'
            # Keep them handy
            shutil.copyfile(affine_mat, os.path.join(self.path, 'affine_from_MNI.mat'))
            #shutil.copyfile(inverse_warp, os.path.join(self.path, 'warp_to_MNI.nii.gz'))
            #shutil.copyfile(warp, os.path.join(self.path, 'warp_from_MNI.nii.gz'))
            # Apply affine transform then warp to put mask in registered space
            out_path = os.path.join(self.path, 'brain_mask.nii.gz')
            ants.apply_transform(mask_path, n4_path, affine_mat, out_path).wait()
            white_out_path = os.path.join(self.path, 'whitematter_mask.nii.gz')
            ants.apply_transform(whitematter_mask_path, n4_path, affine_mat, white_out_path).wait()
            # Save metadata
            self.registration_reference = 'registration_reference.nii.gz'
            self.brain_mask = 'brain_mask.nii.gz'
            self.whitematter_mask = 'whitematter_mask.nii.gz'
            self.save()
            print('done.')

    def process_file(self, input_path, output_path, histogram_reference=None, apply_mask=False):
        '''Process (biascorrect, register, etc.) a single file'''
        # These imports can complain on import, so we'll only get them now.
        print('Processing: ', input_path)
        from pattools import ants
        with TemporaryDirectory() as tmp_dir:
            n4_path = os.path.join(tmp_dir, 'n4.nii')
            ants.n4_bias_correct(input_path, n4_path).wait()

            ref_path = os.path.join(self.path, self.registration_reference)
            out_path = os.path.join(tmp_dir, 'regout.nii')
            ants.affine_registration(n4_path, ref_path, out_path).wait()

            mask = nib.load(os.path.join(self.path, self.brain_mask))
            output = nib.load(out_path)
            outdata = output.get_fdata() * 1 # maybe this will force into an ndarray?
            if apply_mask:
                outdata *= mask.get_fdata()

            output = nib.Nifti1Image(outdata, output.affine, output.header)
            nib.save(output, output_path)

    def normalize_file(self, file, reference_file):
         # normalise whitematter intensity
        if self.whitematter_mask != None and self.brain_mask != None:
            wm_mask = nib.load(os.path.join(self.path, self.whitematter_mask)).get_fdata()
            mask = nib.load(os.path.join(self.path, self.brain_mask)).get_fdata()
            img = nib.load(file)
            img_data = img.get_fdata()
            ref_data = nib.load(reference_file).get_fdata()
            outdata = normalize_by_whitematter(
                img_data * mask,
                ref_data * mask,
                wm_mask)

            output = nib.Nifti1Image(outdata, output.affine, output.header)
            nib.save(output, file)

    def normalize_data(self, filter_name, brain_mask=True, do_znorm=True, do_histogram_match=False):
        '''This will generate a slice-wise normalized set of data (which you'll need to fit in RAM)'''
        print('Normalizing data for filter', filter_name)

        mask = nib.load(os.path.join(self.path, self.brain_mask)).get_fdata() if brain_mask and self.brain_mask != None else 1.
        files = [
            os.path.join(self.path, fm.study_date, fm.processed_file)
            for fm in self.files_for_filter(filter_name)]

        if len(files) == 0:
            print('No files for filter', filter_name)
            return

        # Work out the dimensions for our output
        first_img = nib.load(files[0])
        first_data = first_img.get_fdata()

        # We're going to try to fit the whole thing into memory because we have tonnes of RAM
        # We could refactor this to work via HDD if that's an issue but you'd probably see thrashing.
        # The output ndarray will be coronal, study index (time), then the two other spatial dims (should check).
        output = np.zeros((first_data.shape[0], len(files),first_data.shape[1],first_data.shape[2]))

        # Load all the timeline data into a single array (better have the memory for it)
        for i, file in enumerate(files):
            data = nib.load(file).get_fdata()
            for slice_index in range(data.shape[0]):
                output[slice_index,i,:,:] = data[slice_index,:,:] * mask[slice_index,:,:]
        print(f'Loaded {len(files)} files in {output.size*output.itemsize / (1024.*1024.)}MB of RAM')

        # Background threshold for Z-Norm
        background_threshold = 20
        print('processing', output.shape[0], 'slices')
        for slice_index in range(output.shape[0]):
            print('       ', slice_index, '/', output.shape[0])
            # We're going to take the most recent study as our reference
            bg_mask = output[slice_index,-1,:,:] > background_threshold
            # Skip the background only slices.
            if (np.sum(bg_mask.astype(float)) <= 0):
                continue
            mean0 = np.mean(output[slice_index,-1,:,:][bg_mask])
            std0 = np.std(output[slice_index,-1,:,:][bg_mask])
            if mean0 == 0 or std0 == 0:
                continue
            for i in range(0, output.shape[1]): # Where i is the study index
                bg_mask = output[slice_index,i,:,:] > background_threshold
                if (np.sum(bg_mask.astype(float)) <= 0):
                    continue
                sliceI = output[slice_index,i,:,:]
                if len(sliceI) != 0:
                    meanI = np.mean(output[slice_index,i,:,:][bg_mask])
                    stdI = np.std(output[slice_index,i,:,:][bg_mask])
                    if meanI == 0 or stdI == 0:
                        continue
                    # Apply Z-Norm and Histogram Matching
                    if do_histogram_match:
                        output[slice_index,i,:,:] = histogram_match(output[slice_index,i,:,:], output[slice_index,0,:,:])
                    if do_znorm: 
                        output[slice_index,i,:,:] = (output[slice_index,i,:,:] - meanI) / stdI * std0 + mean0

        # Make sure the minimum is zero (so we don't get grey backgrounds)
        output = output - np.min(output)
        # Norm the data to full uint16 range (this makes export way easier)
        output = output / np.max(output) * np.iinfo(np.uint16).max
        #output = output.astype(np.uint16)
        # Export the data
        print('max, min:', np.max(output), ",", np.min(output))
        for i, f in enumerate(files):
            img = nib.load(f)
            nib.save(nib.Nifti1Image(output[:,i,:,:] * mask, img.affine, img.header), f)
        
        # TODO, this should probably be moved to the renderer
        wl, ww  = estimate_window(output) # Window level, and window width are the convention
        min_val = np.min(output)
        max_val = np.max(output)
        if min_val == np.nan: min_val = 0
        if max_val == np.nan: max_val = 0
        self.normed_window_map[filter_name] = {'min':min_val, 'max':max_val, 'ww': ww, 'wl': wl}
        for k in self.normed_window_map[filter_name].keys():
            if self.normed_window_map[filter_name][k] == np.NaN or self.normed_window_map[filter_name][k] == np.Infinity:
                self.normed_window_map[filter_name][k] = 0
        print(self.normed_window_map)
        self.save()

    def process(self):
        '''Process (bias correct, register to reference, etc.) all image files'''
        if (self.registration_reference == None
            or os.path.exists(os.path.join(self.path, self.registration_reference)) == False
            or self.brain_mask == None
            or os.path.exists(os.path.join(self.path, self.brain_mask)) == False):
            self.setup_registration_reference()

        files_to_process = []
        self._load_datamap()
        # TODO -- Not currently using the given histogram reference, need to either remove or use...
        histogram_references = {}
        # TODO -- We seem to be selecting the most recent study as the histogram match
        # this will break on update
        #for filter in self.filters:
            #histogram_ref_path = os.path.join(self.path, f'intensity_ref_{filter.name}.nii.gz')
            #if not os.path.exists(histogram_ref_path):
            #    candidate_path = ""
            #    for study_date in self.datamap:
            #        for fm in self.datamap[study_date]:
            #            if fm.filter_name == filter.name:
            #                candidate_path = os.path.join(self.path, study_date, fm.processed_file)
            #                if not os.path.exists(candidate_path):
            #                    candidate_path = os.path.join(self.path, study_date, fm.file)
            #    try:
            #        shutil.copyfile(candidate_path, histogram_ref_path)
            #    except:
            #        print('Failed to copy histogram reference ', candidate_path)
            #histogram_references[filter.name] = histogram_ref_path

        for study in self.datamap:
            study_path = os.path.join(self.path, study)
            for filemeta in self.datamap[study]:
                print(os.path.join(study_path, filemeta.processed_file), 'exists?', os.path.exists(os.path.join(study_path, filemeta.processed_file)))
                if os.path.exists(os.path.join(study_path, filemeta.processed_file)) == False and os.path.exists(os.path.join(study_path, filemeta.file)) == True:
                    input = os.path.join(self.path, study, filemeta.file)
                    output = os.path.join(self.path, study, filemeta.processed_file)
                    filter_name = filemeta.filter_name
                    files_to_process.append((input, output))
        # Add a progress bar
        print('Processing', len(files_to_process), 'files...')
        files_to_process = progress.bar(files_to_process, expected_size=len(files_to_process))
        for input, output in files_to_process: # TODO: Discarding histogram_reference
            self.process_file(input, output)

        # Now that all the files are processed (including the histogram references) we can normalize
        for fil in self.filters:
            self.normalize_data(fil.name)

    def study_dates(self):
        '''Returns all study dates, whether we have a scan or not'''
        folders = next(os.walk(self.path))[1]
        return [datetime.strptime(f, '%Y%m%d').date() for f in folders]

    def set_manual_series(self, filter_name, study_date, series):
        '''Sets a manually selected series for a given filter/study combination'''
        self.manual_studies[filter_name, study_date] = series

    def files_for_filter(self, filter_name):
        '''Returns the file metadata objects for a given filtername, sorted by date'''
        result = []
        self.load() # Load datamap from disk
        for study in self.datamap.keys():   
            for filemeta in self.datamap[study]:
                if filemeta.filter_name == filter_name:
                    result.append(filemeta)

        result.sort(key=lambda fm: fm.study_date)
        return result



class ScorecardElement:
    '''The scorecard element is used to create a series filter.'''
    description = None
    points = 0
    def __init__(self, description, points):
        self.description = description
        self.points = points

class Filter:
    '''The filter class extracts a 0 or 1 series from a study (which math the filter, obvs.)'''
    name = None
    scorecard = None # For matching description
    min_rows = 100
    min_cols = 100
    threshold = 0

    def __init__(self, name):
        self.name = name
        self.scorecard = []

    def filter(self, study):
        '''Returns the series which best matches the filter, or None if there are no good matches.'''
        series = study.find_series()
        candidates = []

        for seri in series:
            score = 0
            # If the description contains a string on the scorecard..
            for se in self.scorecard:
                if se.description.lower() in seri.description.lower():
                    score += se.points #... add to the score
            if score >= self.threshold:
                candidates.append((score, seri))
        # Sort the candidates by score
        candidates.sort(key=lambda x: x[0], reverse=True)
        # If theres a candidate, append the best
        if len(candidates) > 0:
            return candidates[0][1]
        # No valid series found
        return None


def flair_filter():
    '''Default filter for FLAIR studies'''
    filter = Filter('FLAIR')
    filter.scorecard.append(ScorecardElement('flair', 100))
    filter.scorecard.append(ScorecardElement('mprage', -100))
    filter.scorecard.append(ScorecardElement('localiser', -50))
    filter.scorecard.append(ScorecardElement('sag', 20))
    filter.scorecard.append(ScorecardElement('_cor', -5))
    filter.scorecard.append(ScorecardElement('_tra', -5))
    filter.scorecard.append(ScorecardElement('_cor', -5))
    filter.scorecard.append(ScorecardElement('t2', 45))
    filter.scorecard.append(ScorecardElement('t1', -45))
    filter.scorecard.append(ScorecardElement('3d', 25))
    filter.scorecard.append(ScorecardElement('dis3d', -25))
    filter.scorecard.append(ScorecardElement('report', -50))
    filter.scorecard.append(ScorecardElement('fused', -35))
    filter.scorecard.append(ScorecardElement('spine', -255))
    
    filter.threshold = 100
    return filter

def mprage_cplus_filter():
    '''Default filter for MPRAGE c+ studies'''
    filter = Filter('MPRAGE_C+')
    filter.scorecard.append(ScorecardElement('mprage', 60))
    filter.scorecard.append(ScorecardElement('flair', -100))
    filter.scorecard.append(ScorecardElement('localiser', -50))
    filter.scorecard.append(ScorecardElement('sag', 20))
    filter.scorecard.append(ScorecardElement('_cor', -5))
    filter.scorecard.append(ScorecardElement('_tra', -5))
    filter.scorecard.append(ScorecardElement('_cor', -5))
    filter.scorecard.append(ScorecardElement('t1', 30))
    filter.scorecard.append(ScorecardElement('t2', -30))
    filter.scorecard.append(ScorecardElement('3d', 25))
    filter.scorecard.append(ScorecardElement('dis3d', -25))
    filter.scorecard.append(ScorecardElement('c+', +50))
    filter.scorecard.append(ScorecardElement('c-', -50))
    filter.scorecard.append(ScorecardElement('report', -50))
    filter.scorecard.append(ScorecardElement('fused', -35))
    filter.scorecard.append(ScorecardElement('tv', -35))
    filter.scorecard.append(ScorecardElement('cv', -35))
    filter.scorecard.append(ScorecardElement('lv', -35))
    filter.scorecard.append(ScorecardElement('devmap_rgb', -35))
    filter.scorecard.append(ScorecardElement('spine', -255))
    filter.threshold = 100
    return filter

def mprage_cminus_filter():
    '''Default filter for MPRAGE C- studies'''
    filter = Filter('MPRAGE_C-')
    filter.scorecard.append(ScorecardElement('mprage', 100))
    filter.scorecard.append(ScorecardElement('flair', -100))
    filter.scorecard.append(ScorecardElement('localiser', -50))
    filter.scorecard.append(ScorecardElement('sag', 20))
    filter.scorecard.append(ScorecardElement('_cor', -5))
    filter.scorecard.append(ScorecardElement('_tra', -5))
    filter.scorecard.append(ScorecardElement('_cor', -5))
    filter.scorecard.append(ScorecardElement('t1', 30))
    filter.scorecard.append(ScorecardElement('t2', -30))
    filter.scorecard.append(ScorecardElement('3d', 25))
    filter.scorecard.append(ScorecardElement('dis3d', -25))
    filter.scorecard.append(ScorecardElement('c+', -50))
    filter.scorecard.append(ScorecardElement('c-', +50))
    filter.scorecard.append(ScorecardElement('report', -50))
    filter.scorecard.append(ScorecardElement('fused', -35))
    filter.scorecard.append(ScorecardElement('tv', -35))
    filter.scorecard.append(ScorecardElement('cv', -35))
    filter.scorecard.append(ScorecardElement('lv', -35))
    filter.scorecard.append(ScorecardElement('devmap_rgb', -35))
    filter.scorecard.append(ScorecardElement('spine', -255))
    filter.threshold = 100
    return filter

def default_filters():
    '''Default filter list'''
    return [flair_filter(), mprage_cplus_filter(), mprage_cminus_filter()]


import os
import subprocess
ANTSPATH = ''
if 'ANTSPATH' not in os.environ:
    print('***************** OH NO! ********************')
    print("Can't find ANTSPATH environment value. D:")
    print("pattools.ants is just a thin wrapper around a couple of ANTS tools that you need to install yourself.")
    print("We'll assume it's here somewhere (and in your PATH), but if I don't work try the docker image.")
    print('***************** /OH NO! *******************')
else:
    ANTSPATH = os.environ['ANTSPATH']

def n4_bias_correct(input, output):
    '''Calls N4BiasFieldCorrection'''
    p = subprocess.Popen([os.path.join(ANTSPATH,'N4BiasFieldCorrection'), '-i', input, '-o', output])
    return p

def affine_registration(floating, fixed, output):
    '''Calls antsRegistration'''
    p = subprocess.Popen([
        os.path.join(ANTSPATH,'antsRegistration'),
        '--dimensionality','3', # Run ANTS on 3 dimensional image
        '--float', '1',
        '--interpolation', 'Linear',
        '--use-histogram-matching', '0',
        '--initial-moving-transform', f'[{fixed},{floating},1]',
        '--transform', 'Affine[0.1]',
        '--metric', f'MI[{fixed},{floating},1,32,Regular,0.25]', # Use mutal information (we're not normalizing intensity)
        '--convergence', '[1000x500x250x100,1e-6,10]',
        '--shrink-factors', '8x4x2x1',
        '--smoothing-sigmas', '3x2x1x0vox',
        '--output', f'[{output}_,{output}]'
    ])
    return p

def syn_registration(floating, fixed, output, brain_mask=None):
    '''Calls antsRegistration'''
    p = subprocess.Popen([
        os.path.join(ANTSPATH,'antsRegistration'),
        '--dimensionality','3', # Run ANTS on 3 dimensional image
        '--float', '1',
        '--interpolation', 'Linear',
        '--use-histogram-matching', '0',
        '--initial-moving-transform', f'[{fixed},{floating},1]',
        '--transform', 'Affine[0.1]',
        '--metric', f'MI[{fixed},{floating},1,32,Regular,0.25]', # Use mutal information (we're not normalizing intensity)
        '--convergence', '[1000x500x250x100,1e-6,10]',
        '--shrink-factors', '8x4x2x1',
        '--smoothing-sigmas', '3x2x1x0vox',
        '--transform', 'SyN[0.1,3,0]',
        '--metric', f'CC[{fixed},{floating},1,4]',
        '--convergence', '[100x70x50x20,1e-6,10]',
        '--shrink-factors', '8x4x2x1',
        '--smoothing-sigmas', '3x2x1x0vox',
        '--output', f'[{output}_,{output}]'
    ])
    return p

def apply_transform(floating, fixed, affine, output, warp=None, invert_affine=False):
    '''Calls antsApplyTransforms'''
    # Basic command
    cmd = [
        os.path.join(ANTSPATH,'antsApplyTransforms'),
        '-d', '3',
        '--float', '1',
        '-i', floating,
        '-r', fixed,
        '-n', 'Linear']
    # Invert affine for inverse transforms
    if invert_affine:
        cmd.append('-t')
        cmd.append(f'[{affine},1]')
    else:
        cmd.append('-t')
        cmd.append(affine)
    # Apply warp if included
    if warp != None:
        cmd.append('-t')
        cmd.append(warp)
    # Output
    cmd.append('-o')
    cmd.append(output)

    p = subprocess.Popen(cmd)
    return p



'''The PACs module contains classes that enable simple, patient-based access to PACS data.'''

import ast
import os
from dicom2nifti.convert_dicom import dicom_array_to_nifti
from socket import gethostname
from pydicom.dataset import Dataset
from pynetdicom import (
    AE, evt, build_role,
    PYNETDICOM_IMPLEMENTATION_UID,
    PYNETDICOM_IMPLEMENTATION_VERSION)
# pylint: disable=no-name-in-module
# (these are generated at runtime for this module)
from pynetdicom.sop_class import (
    PatientRootQueryRetrieveInformationModelFind,
    StudyRootQueryRetrieveInformationModelFind,
    StudyRootQueryRetrieveInformationModelGet,
    # Storage classes
    ComputedRadiographyImageStorage,
    DigitalXRayImagePresentationStorage,
    DigitalXRayImageProcessingStorage,
    DigitalMammographyXRayImagePresentationStorage,
    DigitalMammographyXRayImageProcessingStorage,
    CTImageStorage,
    EnhancedCTImageStorage,
    LegacyConvertedEnhancedCTImageStorage,
    UltrasoundMultiframeImageStorage,
    MRImageStorage,
    EnhancedMRImageStorage,
    MRSpectroscopyStorage,
    EnhancedMRColorImageStorage,
    LegacyConvertedEnhancedMRImageStorage,
    UltrasoundImageStorage,
    GrayscaleSoftcopyPresentationStateStorage,
    ColorSoftcopyPresentationStateStorage,
    PseudocolorSoftcopyPresentationStageStorage,
    BlendingSoftcopyPresentationStateStorage,
    XAXRFGrayscaleSoftcopyPresentationStateStorage,
    GrayscalePlanarMPRVolumetricPresentationStateStorage,
    CompositingPlanarMPRVolumetricPresentationStateStorage,
    VolumeRenderingVolumetricPresentationStateStorage,
    SegmentedVolumeRenderingVolumetricPresentationStateStorage,
    MultipleVolumeRenderingVolumetricPresentationStateStorage,
    XRayAngiographicImageStorage,
    EnhancedXAImageStorage,
    XRayRadiofluoroscopicImageStorage,
    EnhancedXRFImageStorage,
    XRay3DAngiographicImageStorage,
    XRay3DCraniofacialImageStorage,
    BreastTomosynthesisImageStorage,
    BreastProjectionXRayImagePresentationStorage,
    BreastProjectionXRayImageProcessingStorage,
    IntravascularOpticalCoherenceTomographyImagePresentationStorage,
    IntravascularOpticalCoherenceTomographyImageProcessingStorage,
    NuclearMedicineImageStorage,
    ParametricMapStorage,
    RawDataStorage,
    SpatialRegistrationStorage,
    SpatialFiducialsStorage,
    DeformableSpatialRegistrationStorage,
    SegmentationStorage,
    SurfaceSegmentationStorage,
    TractographyResultsStorage,
    RealWorldValueMappingStorage,
    SurfaceScanMeshStorage,
    SurfaceScanPointCloudStorage,
    VLEndoscopicImageStorage,
    VideoEndoscopicImageStorage,
    VLMicroscopicImageStorage,
    VideoMicroscopicImageStorage,
    VLSlideCoordinatesMicroscopicImageStorage,
    VLPhotographicImageStorage,
    VideoPhotographicImageStorage,
    OphthalmicPhotography8BitImageStorage,
    OphthalmicPhotography16BitImageStorage,
    StereometricRelationshipStorage,
    OphthalmicTomographyImageStorage,
    WideFieldOphthalmicPhotographyStereographicProjectionImageStorage,
    WideFieldOphthalmicPhotography3DCoordinatesImageStorage,
    OphthalmicOpticalCoherenceTomographyEnFaceImageStorage,
    OphthlamicOpticalCoherenceTomographyBScanVolumeAnalysisStorage,
    VLWholeSlideMicroscopyImageStorage,
    LensometryMeasurementsStorage,
    AutorefractionMeasurementsStorage,
    KeratometryMeasurementsStorage,
    SubjectiveRefractionMeasurementsStorage,
    VisualAcuityMeasurementsStorage,
    SpectaclePrescriptionReportStorage,
    OphthalmicAxialMeasurementsStorage,
    IntraocularLensCalculationsStorage,
    MacularGridThicknessAndVolumeReport,
    OphthalmicVisualFieldStaticPerimetryMeasurementsStorage,
    OphthalmicThicknessMapStorage,
    CornealTopographyMapStorage,
    BasicTextSRStorage,
    EnhancedSRStorage,
    ComprehensiveSRStorage,
    Comprehensive3DSRStorage,
    ExtensibleSRStorage,
    ProcedureSRStorage,
    MammographyCADSRStorage,
    KeyObjectSelectionStorage,
    ChestCADSRStorage,
    XRayRadiationDoseSRStorage,
    RadiopharmaceuticalRadiationDoseSRStorage,
    ColonCADSRStorage,
    AcquisitionContextSRStorage,
    SimplifiedAdultEchoSRStorage,
    PositronEmissionTomographyImageStorage,
    EnhancedPETImageStorage,
    LegacyConvertedEnhancedPETImageStorage,
    BasicStructuredDisplayStorage,
    RTImageStorage
    )


def cget_series(scp_settings, dataset):
    '''Runs a series level CGET request'''
    ae = AE(scp_settings.local_aet)
    # Add the requested presentation contexts (QR SCU)
    ae.add_requested_context(StudyRootQueryRetrieveInformationModelGet)
    # Add the requested presentation context (Storage SCP)
    ae.add_requested_context(CTImageStorage)
    ae.add_requested_context(EnhancedCTImageStorage)
    ae.add_requested_context(MRImageStorage)
    ae.add_requested_context(MRSpectroscopyStorage)
    ae.add_requested_context(EnhancedMRColorImageStorage)
    # Note, we can add to these if we want other types of image (see massive import statement)

    # Create an SCP/SCU Role Selection Negotiation item for Image Storage
    roles = []
    roles.append(build_role(CTImageStorage, scp_role=True))
    roles.append(build_role(EnhancedCTImageStorage, scp_role=True))
    roles.append(build_role(MRImageStorage, scp_role=True))
    roles.append(build_role(MRSpectroscopyStorage, scp_role=True))
    roles.append(build_role(EnhancedMRColorImageStorage, scp_role=True))

    dicoms = []
    def handle_store_series(e):
        nonlocal dicoms

        ds = e.dataset
        context = e.context
        meta = Dataset()
        # Add the DICOM File Meta Information
        meta = Dataset()
        meta.MediaStorageSOPClassUID = ds.SOPClassUID
        meta.MediaStorageSOPInstanceUID = ds.SOPInstanceUID
        meta.ImplementationClassUID = PYNETDICOM_IMPLEMENTATION_UID
        meta.ImplementationVersionName = PYNETDICOM_IMPLEMENTATION_VERSION
        meta.TransferSyntaxUID = context.transfer_syntax

        # Add the file meta to the dataset
        ds.file_meta = meta

        # Set the transfer syntax attributes of the dataset
        ds.is_little_endian = context.transfer_syntax.is_little_endian
        ds.is_implicit_VR = context.transfer_syntax.is_implicit_VR

        dicoms.append(ds)

        return 0x0000

    handlers = [(evt.EVT_C_STORE, handle_store_series)]

    # Associate with peer AE
    assoc = ae.associate(
        scp_settings.host, scp_settings.port,
        ae_title=scp_settings.ae_title, ext_neg=roles, evt_handlers=handlers)

    # We're (apparently) using the handle_store_series 
    if assoc.is_established:
        responses = assoc.send_c_get(dataset, StudyRootQueryRetrieveInformationModelGet)
        for (status, _) in responses:
            if status:
                pass
            else:
                raise TimeoutError('Connection timed out, was aborted or received invalid response')
        assoc.release()
        return dicoms
    else:
        raise ConnectionError('Association rejected, aborted or never connected')

def cget_report(scp_settings, dataset):
    '''Runs a CGET request for a report'''
    ae = AE(scp_settings.local_aet)
    # Add the requested presentation contexts (QR SCU)
    ae.add_requested_context(StudyRootQueryRetrieveInformationModelGet)
    # Add the requested presentation context (Storage SCP)
    ae.add_requested_context(BasicTextSRStorage)
    ae.add_requested_context(EnhancedSRStorage)

    # Create an SCP/SCU Role Selection Negotiation item for CT Image Storage
    roles = []
    roles.append(build_role(BasicTextSRStorage, scp_role=True))
    roles.append(build_role(EnhancedSRStorage, scp_role=True))

    reportds = None
    def handle_store_report(e):
        nonlocal reportds
        reportds = e.dataset
        return  0x0000

    handlers = [(evt.EVT_C_STORE, handle_store_report)]

    # Associate with peer AE
    assoc = ae.associate(
        scp_settings.host, scp_settings.port,
        ae_title=scp_settings.ae_title, ext_neg=roles, evt_handlers=handlers)

    if assoc.is_established:
        responses = assoc.send_c_get(dataset, StudyRootQueryRetrieveInformationModelGet)
        for (status, _) in responses:
            if status:
                pass
            else:
                raise TimeoutError('Connection timed out, was aborted or received invalid response')

        assoc.release()
        return reportds

    else:
        raise ConnectionError('Association rejected, aborted or never connected')

def cfind(scp_settings, dataset, query_model='P'):
    ''' Make a CFIND request '''
    ae = AE(scp_settings.local_aet)
    ae.add_requested_context(StudyRootQueryRetrieveInformationModelFind)
    ae.add_requested_context(PatientRootQueryRetrieveInformationModelFind)
    ae.add_requested_context('1.2.840.10008.5.1.4.1.1.88.11')
    assoc = ae.associate(
        scp_settings.host, scp_settings.port, ae_title=scp_settings.ae_title)
    # I'm sure send_c_find used to accept the single letter versions but it doesn't now.
    if query_model == 'S': query_model = StudyRootQueryRetrieveInformationModelFind
    elif query_model == 'P': query_model = PatientRootQueryRetrieveInformationModelFind

    #Check for successful association
    if assoc.is_established:
         # Use the C-FIND service to send the identifier
         responses = assoc.send_c_find(dataset, query_model=query_model)

         result = []
         for (status, identifier) in responses:
             if status:
                 # If the status is 'Pending' then identifier is the C-FIND response
                 if status.Status in (0xFF00, 0xFF01):
                     result.append(str(identifier))
             else:
                 assoc.release()
                 raise TimeoutError('Connection timed out, was aborted or received invalid response ' + scp_settings.ae_title)

         # Release the association
         assoc.release()
         return result
    else:
        raise ConnectionError('Association rejected, aborted or never connected ' + scp_settings.ae_title)

def cstore(scp_settings, dataset):
    ''' Make a CFIND request '''
    # Initialise the Application Entity
    ae = AE()
    # Add a requested presentation context
    ae.add_requested_context(CTImageStorage)
    ae.add_requested_context(EnhancedCTImageStorage)
    ae.add_requested_context(MRImageStorage)
    ae.add_requested_context(MRSpectroscopyStorage)
    ae.add_requested_context(EnhancedMRColorImageStorage)
    # Create association to scp
    assoc = ae.associate(
        scp_settings.host, scp_settings.port, ae_title=scp_settings.ae_title)
    # Store the image
    if assoc.is_established:
        status = assoc.send_c_store(ds)
        if status:
            pass
        else:
            raise TimeoutError('Connection timed out, was aborted or received invalid response ' + scp_settings.ae_title)

        assoc.release()
    else:
        raise ConnectionError('Association rejected, aborted or never connected ' + scp_settings.ae_title)


def find_studies_from_patient(patient):
    '''Runs a CFIND for all studies for a given patient'''
    ds = Dataset()
    ds.PatientID = patient.id
    ds.PatientName = patient.name
    ds.StudyInstanceUID = ''
    ds.AccessionNumber = ''
    ds.QueryRetrieveLevel = 'STUDY'
    ds.SeriesUID=''
    ds.PatientSize = ''
    ds.ModalitiesInStudy = ''
    ds.SOPClassesInStudy = ''
    ds.StudyDescription = '*'
    ds.FailedSOPSequence = ''
    ds.DerivationCodeSequence =''
    ds.SeriesInStudy=''
    ds.StudyDate=''
    ds.StudyTime=''
    results = cfind(patient.scp_settings, ds)
    studies = []
    for result in results:
        study = Study.parse_result(result)
        study.scp_settings = patient.scp_settings
        if study != None: studies.append(study)

    return studies

def find_series_from_study(study):
    '''CFINDS all series for a given study'''
    ds = Dataset()
    ds.StudyInstanceUID = study.study_uid
    ds.AccessionNumber = ''
    ds.QueryRetrieveLevel = 'SERIES'
    ds.SeriesInstanceUID=''
    ds.SeriesDescription = ''
    ds.SeriesNumber = ''
    ds.Modality =''
    ds.SeriesDate=''
    ds.StudyDate = ''
    ds.SeriesTime=''
    ds.StudyDescription = '*'

    results = cfind(study.scp_settings, ds, query_model='S')
    series = []
    for result in results:
        seri = Series.parse_result(result)
        seri.scp_settings = study.scp_settings
        if seri != None: series.append(seri)

    return series




#def find_studies(scp_settings, patient_name = '', patient_id = '' ,study_uid = '',
#    accession_number = '', modalities_in_study = '', study_date = '',
#    study_time = '', description = ''):
#    '''Run a study level CFIND requests using the supplied parameters'''
#    ds = Dataset()
#    ds.PatientID = patient_id
#    ds.PatientName = patient_name
#    ds.StudyInstanceUID = study_uid
#    ds.AccessionNumber = accession_number
#    ds.QueryRetrieveLevel = 'STUDY'
#    ds.ModalitiesInStudy = str(modalities_in_study)
#    ds.StudyDescription = description
#    ds.SeriesInStudy=''
#    ds.StudyDate=study_date
#    ds.StudyTime=study_time
#
#    results = cfind(scp_settings, ds)
#    studies = []
#    for result in results:
#        study = Study.parse_result(result)
#        study.scp_settings = patient.scp_settings
#        if study != None: studies.append(study)
#
#    return studies

### ScpSettings class
class ScpSettings:
  '''A class representing SCP settings (i.e. how to connect to PACS)'''
  ae_title = None
  host = None
  port = None
  local_aet = None

  def __init__(self, ae_title, host, port, local_aet=None):
      self.ae_title = ae_title
      self.host = host
      self.port = port
      if local_aet != None: self.local_aet = local_aet
      else: self.local_aet = ae_title
      # The above is a little sneaky because the remote PACS might not want to
      # talk to an unknown SCP. But it knows itself, so it will often be happy
      # to dump all its info to someone with the same name.

  def __str__(self):
      return (
        'ae_title[' + str(self.ae_title) +
        '], host[' + str(self.host) +
        '], port[' + str(self.port) + ']'
      )

### Patient class
class Patient:
    ''' Contains the details for a single patient '''
    id = None
    name = ''
    dob = ''
    patient_sex = ''
    scp_settings = None

    def __init__(self, id=None, scp_settings=None):
        ''' Requires an id string and the scp_settings of the remote PACs scp '''
        self.id = id
        self.scp_settings = scp_settings
        if id != None and scp_settings !=None:
            patient = Patient.from_id(id, scp_settings)
            if patient != None:
                self.name = patient.name
                self.dob = patient.dob
                self.patient_sex = patient.patient_sex
            else:
                print(f'Patient #{id} was not found in SCP')

    def __str__(self):
        return (
            'id          : ' + str(self.id) + '\n' +
            'name        : ' + str(self.name) + '\n' +
            'dob         : ' + str(self.dob) + '\n' +
            'patient_sex : ' + str(self.patient_sex) + '\n' +
            'scp_settings: ' + str(self.scp_settings))

    def find_studies(self):
        ''' Find all remote studies for the patient '''
        return find_studies_from_patient(self)

    @staticmethod
    def from_id(id, scp_settings):
        '''Returns a patient from a given patient id'''
        ds = Dataset()
        ds.PatientID = id
        ds.PatientName = ''
        ds.PatientBirthDate = ''
        ds.PatientSex = ''
        ds.QueryRetrieveLevel = 'PATIENT'
        results = cfind(scp_settings, ds)
        if len(results) > 0:
            return Patient.parse_result(results[0], scp_settings)
        else:
            return None

    @staticmethod
    def parse_result(result, scp_settings):
        ''' Parse a pynetdicom result and return a patient '''
        patient = Patient(scp_settings=scp_settings)
        for field in result.split('\n'):
            if field.startswith('(0010, 0010)'):
                patient.name = field.split()[-1].strip("'").strip('"')
            elif field.startswith('(0010, 0020)'):
                patient.id = field.split()[-1].strip("'").strip('"')
            elif field.startswith('(0010, 0040)'):
                patient.patient_sex = field.split()[-1].strip("'").strip('"')
            elif field.startswith('(0010, 0030)'):
                patient.dob = field.split()[-1].strip("'").strip('"')
        return patient
        
### Study class
class Study:
    ''' Contains the details for a single study'''
    study_uid = None
    patient_id = None
    accession_number = None
    modalities_in_study = None
    study_date = None
    study_time = None
    description = None
    scp_settings = None

    def __init__(self, scp_settings=None):
        self.scp_settings = scp_settings

    def __str__(self):
        return (
            'study_uid           : ' + str(self.study_uid) + '\n' +
            'accession_number    : ' + str(self.accession_number) + '\n' +
            'modalities_in_study : ' + str(self.modalities_in_study) + '\n' +
            'study_date          : ' + str(self.study_date) + '\n' +
            'study_time          : ' + str(self.study_time) + '\n' +
            'description         : ' + str(self.description)+ '\n' +
            'scp_settings        : ' + str(self.scp_settings))

    def find_series(self):
        '''Find all series for this study'''
        return find_series_from_study(self)

    def get_report(self):
        '''CGET the report for this study'''
        series = self.find_series()
        for seri in [s for s in series if s.modality == 'SR']:
            if seri.modality == 'SR':
                ds = Dataset()
                ds.SeriesInstanceUID = seri.series_uid
                ds.StudyInstanceUID = self.study_uid
                ds.QueryRetrieveLevel = 'SERIES'
                return Report(cget_report(self.scp_settings, ds))

                #for result in results:
                #    print("Result:::::::: ", result)
                #if (len(results) > 0):
                #    report = Report.parse_result(results[0])
                #    report.study = self
            print('could not find report')
            return None

    @staticmethod
    def parse_result(result):
        ''' Parse a pynetdicom result and return a study '''
        study = Study()
        for field in result.split('\n'):
            if field.startswith('(0008, 0020)'):
                study.study_date = field.split()[-1].strip("'").strip('"')
            elif field.startswith('(0008, 0030)'):
                study.study_time = field.split()[-1].strip("'").strip('"')
            elif field.startswith('(0008, 0050)'):
                study.accession_number = field.split(': ')[-1].strip("'").strip('"')
            elif field.startswith('(0008, 1030)'):
                study.description = field.split(': ')[-1].strip("'").strip('"')
            elif field.startswith('(0008, 0061)'):
                try:
                    study.modalities_in_study = ast.literal_eval(field.split(': ')[-1].strip("'").strip('"'))
                except:
                    study.modalities_in_study = [field.split(': ')[-1].strip("'").strip('"')]
            elif field.startswith('(0020, 000d)'):
                study.study_uid = field.split()[-1].strip("'").strip('"')
            elif field.startswith('(0010, 0020)'):
                study.patient_id = field.split()[-1].strip("'").strip('"')

        return study

class Series:
    '''Contains the details of a single series'''
    series_uid = None
    study_uid = None
    modality = None
    series_number = None
    description = None
    study_date = None
    scp_settings = None

    def __init__(self, scp_settings=None):
        self.scp_settings = scp_settings

    def __str__(self):
        return (
            'series_uid   : ' + str(self.series_uid) + '\n' +
            'study_uid     : ' + str(self.study_uid) + '\n' +
            'modality     : ' + str(self.modality) + '\n' +
            'series_number: ' + str(self.series_number) + '\n' +
            'description  : ' + str(self.description)+ '\n' +
            'study_date  : ' + str(self.study_date)+ '\n' +
            'scp_settings : ' + str(self.scp_settings))

    def save_dicom(self, dir_path):
        '''Save this series as a dicom (to folder)'''
        if not os.path.exists(dir_path): os.mkdir(dir_path)
        ds = Dataset()
        ds.SeriesInstanceUID = self.series_uid
        ds.StudyInstanceUID = self.study_uid
        ds.QueryRetrieveLevel = 'SERIES'
        dicoms = cget_series(self.scp_settings, ds)
        for dicom in dicoms:
             dicom.save_as(os.path.join(dir_path, dicom.SOPInstanceUID +'.dcm'), write_like_original=False)
        return True

    def save_nifti(self, file_path, max_attempts=5):
        '''Save this series as a nifti file'''
        print('saving file:', file_path)
        if not os.path.exists(os.path.dirname(file_path)) and os.path.dirname(file_path) != '':
            os.mkdir(os.path.dirname(file_path))
        ds = Dataset()
        ds.SeriesInstanceUID = self.series_uid
        ds.StudyInstanceUID = self.study_uid
        ds.QueryRetrieveLevel = 'SERIES'
        dicoms = None
        attempts = 0
        while dicoms == None and attempts < max_attempts:
            print('trying c_get')
            attempts += 1
            try:
                dicoms = cget_series(self.scp_settings, ds)
            except Exception as e:
                print(e)
                if attempts >= max_attempts:
                    raise
        try:
            dicom_array_to_nifti(dicoms, file_path, reorient_nifti=True)
            return True
        except:
            print('Failed to convert series to nifti, maybe try DICOM?')
            return False

    @staticmethod
    def parse_result(result):
        ''' Parse a pynetdicom result and return a series '''
        series = Series()
        for field in result.split('\n'):
            if field.startswith('(0020, 000e)'):
                series.series_uid = field.split()[-1].strip("'").strip('"')
            elif field.startswith('(0008, 0060)'):
                series.modality = field.split()[-1].strip("'").strip('"')
            elif field.startswith('(0020, 0011)'):
                series.series_number = field.split(': ')[-1].strip("'").strip('"')
            elif field.startswith('(0008, 103e)'):
                series.description = field.split(': ')[-1].strip("'").strip('"')
            elif field.startswith('(0020, 000d)'):
                series.study_uid = field.split()[-1].strip("'").strip('"')
            elif field.startswith('(0008, 0020)'):
                series.study_date = field.split(': ')[-1].strip("'").strip('"')
        return series

    @staticmethod
    def from_series_uid(uid, scp_settings):
        '''Get a series from the unique idenifier string'''
        ds = Dataset()
        ds.SeriesUID = uid
        ds.QueryRetrieveLevel = 'SERIES'
        results = cfind(scp_settings, ds, query_model='S')
        if len(results) == 0:
            return None
        return Series.parse_result(results[0])

class Report(Series):
    '''This class represents a study report'''
    patient_name = None
    patient_id = None
    patient_dob = None
    patient_sex = None
    study_uid = None
    series_uid = None
    completion_flag = None
    verification_flag = None
    type = None
    text = None
    json_string = None

    def __init__(self, dataset=None, study=None):
            self.study = study
            if dataset != None:
                self.patient_name = dataset.PatientName
                self.patient_id = dataset.PatientID
                self.patient_dob = dataset.PatientBirthDate
                self.patient_sex = dataset.PatientSex
                self.study_uid = dataset.StudyInstanceUID
                self.series_uid = dataset.SeriesInstanceUID
                self.completion_flag = dataset.CompletionFlag
                self.verification_flag = dataset.VerificationFlag
                self.type = dataset.ContentSequence[0].ValueType
                # Sometimes they're a zip file or something and that's a big fat TODO
                self.json_string = dataset.to_json()


    def __str__(self):
        return (
            'REPORT: ' + '\n' +
            'patient_name     : ' + str(self.patient_name) + '\n' +
            'patient_id       : ' + str(self.patient_id) + '\n' +
            'patient_dob      : ' + str(self.patient_dob) + '\n' +
            'patient_sex      : ' + str(self.patient_sex)+ '\n' +
            'study_uid        : ' + str(self.study_uid)+ '\n' +
            'series_uid       : ' + str(self.series_uid)+ '\n' +
            'completion_flag  : ' + str(self.completion_flag)+ '\n' +
            'verification_flag: ' + str(self.verification_flag)+ '\n' +
            'type             : ' + str(self.type)+ '\n' +
            'text             : ' + str(self.text)+ '\n' +
            'json_string      : ' + str(self.json_string))

    @staticmethod
    def parse_result(result):
        ''' Parse a pynetdicom result and return a report '''
        report = Report()
        for field in result.split('\n'):
            if field.startswith('(0040, a040)'):
                report.type = field.split(': ')[-1].strip("'").strip('"')
            elif field.startswith('(0040, a160)'):
                report.text = field.split(': ')[-1].strip("'").strip('"')
            elif field.startswith('(0040, a491)'):
                report.series_date = field.split(': ')[-1].strip("'").strip('"')
        return report



import sys
import numpy as np

def histogram_match(image, reference):
    if 'skimage.exposure' not in sys.modules:
        from skimage.exposure import match_histograms
    '''Returns image modified to match reference's histogram'''
    return match_histograms(image, reference, multichannel=False)

def _build_gaussian_mixture(data, n_components, random_sample=100000):
    '''Builds and fits a GaussianMixture from a random sample of the provided data'''
    if 'skimage.mixture' not in sys.modules:
        from sklearn.mixture import GaussianMixture
    gm = GaussianMixture(n_components=n_components, random_state=42069)
    rng = np.random.default_rng(42069) # We're seeding our RNG to make it deterministic
    gm.fit(rng.choice(data.reshape(-1), random_sample).reshape(-1,1))
    return gm

def _find_key_intesities(data, mixture_model):
    ''' Finds the key intensity values for scaling based on the mixture model.'''
    max_val = np.max(data)
    if max_val == 0:
        return [0, 0]
    dist = mixture_model.predict_proba(np.arange(max_val).reshape(-1,1))
    # Find the ordered distribution indexes
    ordered_idx = []
    idx = -1
    for d in dist:
        imax = np.argmax(d)
        if idx != imax:
            idx = imax
            ordered_idx.append(idx)
    # Find the key intensity values
    keys = []
    # Minimum value is the first key (should usually be zero).
    keys.append(np.min(data))
    # For each distribution within the model where oi is the index into 
    # ordered idx and di is the index into the dist entries.
    # We're ignoring the last distribution because the peak is poorly defined
    # and there's no next crossover point.
    for oi, di in enumerate(ordered_idx[:-1]):
        # Add the peak value
        peak = np.argmax(dist[:,di])
        keys.append(peak)
        # Find the cross over point
        # Check for the first value from the peak where the next distribution
        # becomes more probable than the current one. Add this as a key point.
        di_next = ordered_idx[oi+1]
        i = peak
        while (i < max_val and di_next < len(dist[i])):
            if dist[i][di] <= dist[i][di_next]:
                keys.append(i)
                break
            i+=1
    # Finally add the max value and return.
    keys.append(max_val)
    return np.array(keys)

def mixed_model_match(image, reference, n_components=5):
    # Build gaussian mixture models
    ref_gm = _build_gaussian_mixture(reference, n_components)
    img_gm = _build_gaussian_mixture(image, n_components)
    # Find the key intensity values for scaling
    ref_keys = _find_key_intesities(reference, ref_gm)
    img_keys = _find_key_intesities(image, img_gm)

    # We're going to use bias as the minimum value and scale as max
    # This will mean that we should end up with the same min and max vals
    #ref_bias = ref_keys[0]
    #ref_scale = ref_keys[-1] - ref_bias
    #img_bias = img_keys[0]
    #img_scale = img_keys[-1] - img_bias
    ## Apply the operation to both the image and the keys
    #image = (image - img_bias) / img_scale * ref_scale + ref_bias
    #img_keys = (img_keys - img_bias) / img_scale * ref_scale + ref_bias
    # Get a set of intervals
    img_intervals = list(zip(img_keys, img_keys[1:]))
    ref_intervals = list(zip(ref_keys, ref_keys[1:]))
    # We want to calculate the masks before we change thing.
    # Although it will be a monotocally increasing set there may be overlap in the process.
    interval_masks = []
    for istart, iend in img_intervals:    
        interval_masks.append(np.logical_and(image > istart, image <= iend))
    interval_masks.reverse() #this allows us to pop()
    # Now we scale each interval individually
    for (istart, iend), (rstart, rend) in zip(img_intervals, ref_intervals):
        interval_mask = interval_masks.pop()
        rbias = rstart
        rscale = rend - rstart
        ibias = istart
        iscale = iend - istart
        image[interval_mask] -= ibias
        print('ibias, rbias',ibias, rbias)
        if iscale != 0:
            image[interval_mask] /= iscale
        image[interval_mask] = image[interval_mask] * rscale + rbias
    # ...and done!
    return image

def vistarsier_compare(c, p, min_val=-1., max_val=5., min_change=0.8, max_change=3.):
    """ VisTarsier's compare operation
    Parameters
    ----------
    c : ndarray
        The current volume
    p : ndarray
        The prior volume
    min_val : float
        The minimum value (measured in standard deviations) to consider
    max_val : float
        The maximum value (measured in standard deviations) to consider
    min_change : float
        The minimum change of value (measured in standard deviations) to consider
    max_change : float
        The maximum change of value (measured in standard deviations) to consider
    Returns
    -------
    change : ndarray
        The relevant change in signal.
    """
    # Get standard deviations for current and prior
    pstd = p.std()
    cstd = c.std()
    # Align prior standard deviation to current
    p = ((p - p.mean()) / pstd) * cstd + c.mean()

    #Calculate change
    change = c - p
    # Ignore change outside of minimuim and maximum values
    change[c < min_val*cstd] = 0
    change[p < min_val*cstd] = 0
    change[c > max_val*cstd] = 0
    change[p > max_val*cstd] = 0
    change[np.abs(change) < min_change*cstd] = 0
    change[np.abs(change) > max_change*cstd] = 0

    return change

def norm_z_value(img, ref_img):
    """This function will normalize the two images using z-value normalization
    Parameters
    ----------
    img : ndarray
        The image to normalize
    ref_img : ndarray
        The referene image
    Returns
    -------
    img : ndarray
        The image that's been normalized.
    """
    # Get standard deviations for current and prior
    imgstd = img.std()
    refstd = ref_img.std()
    # Align prior standard deviation to current
    img = ((img - img.mean()) / imgstd) * refstd + ref_img.mean()

    return img

def normalize_by_whitematter(img, ref_img, white_matter_mask):
    """This function will normalize two MRI brain images by histogram matching
    followed by z-value normilization on the whitematter based on a given mask.
    Parameters
    ----------
    img : ndarray
        The image to normalize
    ref_img : ndarray
        The referene image
    white_matter_mask : ndarray
        Mask where 1 in white matter and 0 is non-white matter.
    Returns
    -------
    img : ndarray
        The image that's been normalized.
    """
    # First we histogram match the whole image
    img = histogram_match(img, ref_img)
    # Then we're going to perform z-score normalisation using the whitematter
    # masked means and std deviation. This should get the whitematter values
    # as close as possible.
    masked_ref = ref_img * white_matter_mask
    masked_img = img * white_matter_mask
    mrstd = masked_ref.std()
    mistd = masked_img.std()
    normed_img = ((img - masked_img.mean()) / mistd) * mrstd + masked_ref.mean()

    return normed_img

def estimate_window(data: np.ndarray):
    '''Estimates a viewing window based on the given numpy array. Just using mean and std but it's a start'''
    wl = np.mean(data[data > 20])
    ww = 3 * np.std(data[data > 20])
    
    if wl == np.nan or wl == np.Infinity: wl = 0
    if ww == np.nan or ww == np.Infinity: ww = 0
    return (wl, ww)

# This function was stolen from a numpy cookbook and provides smoothing as is 
# extrememly well commented (and then expressed through some very dense code with single letter vars). 
def smooth(x, window_len=11, window='hanning'):
    """smooth the data using a window with requested size.
    This method is based on the convolution of a scaled window with the signal.
    The signal is prepared by introducing reflected copies of the signal 
    (with the window size) in both ends so that transient parts are minimized
    in the begining and end part of the output signal.
    input:
        x: the input signal 
        window_len: the dimension of the smoothing window; should be an odd integer
        window: the type of window from 'flat', 'hanning', 'hamming', 'bartlett', 'blackman'
            flat window will produce a moving average smoothing.interpolate
    example:
    t=linspace(-2,2,0.1)
    x=sin(t)+randn(len(t))*0.1
    y=smooth(x)
    see also: 
    numpy.hanning, numpy.hamming, numpy.bartlett, numpy.blackman, numpy.convolve
    scipy.signal.lfilter
    TODO: the window parameter could be the window itself if an array instead of a string
    NOTE: length(output) != length(input), to correct this: return y[(window_len/2-1):-(window_len/2)] instead of just y.
    """
    if x.ndim != 1:
        raise ValueError("smooth only accepts 1 dimension arrays.")
    if x.size < window_len:
        raise ValueError("Input vector needs to be bigger than window size.")

    if window_len<3:
        return x

    if not window in ['flat', 'hanning', 'hamming', 'bartlett', 'blackman']:
        raise ValueError("Window is on of 'flat', 'hanning', 'hamming', 'bartlett', 'blackman'")

    s=np.r_[x[window_len-1:0:-1],x,x[-2:-window_len-1:-1]]
    #print(len(s))
    if window == 'flat': #moving average
        w=np.ones(window_len,'d')
    else:
        w=eval('np.'+window+'(window_len)')
    y=np.convolve(w/w.sum(),s,mode='valid')
    return y
	
def m_mode(img, start_point, end_point, out_size=100):
    ''' This function produces an interpolated 1-D view of the line between 2 points '''
    if 'scipy.interpolate' not in sys.modules:
        import scipy.interpolate
    if len(img.shape) != 3:
        raise ValueError('m_mode only works on 3D volumes')
    # This will give us the points in real numbers
    xs = np.linspace(start_point[0], end_point[0], out_size)
    ys = np.linspace(start_point[1], end_point[1], out_size)
    zs = np.linspace(start_point[2], end_point[2], out_size)

    # This will give us the grid points for the data (ugly but i.shape[1] w + list(range(img.shape[2]o))rks)
    points = [list(range(img.shape[0])), list(range(img.shape[1])), list(range(img.shape[2]))]
    outdata = scipy.interpolate.interpn(points, img, (xs,ys,zs))
    return outdata

def rgbstuffed_to_16bit(rbgstuffed_array):
    sixteenbit_out = np.zeros((rbgstuffed_array.shape[0], rbgstuffed_array.shape[1]), np.uint16)
    sixteenbit_out[:,:] = (rbgstuffed_array[:,:,1] << 8).astype(np.uint16) + rbgstuffed_array[:,:,2]
    return sixteenbit_out

def sixteenbit_to_rgbstuffed(sixteenbit_array):
    rgbout = np.zeros((sixteenbit_array.shape[0], sixteenbit_array.shape[1], 3), np.uint8)
    rgbout[:,:,1] = (sixteenbit_array >> 8).astype(np.uint8)
    rgbout[:,:,2] = (sixteenbit_array - (rgbout[:,:,1] << 8)).astype(np.uint8)
    return rgbout


'''A set of helper methods to access handy resources (e.g. Atlases or datasets)'''

from zipfile import ZipFile
import requests
import io
import os
import nibabel as nib
from clint.textui import progress

def download_file(url, file_path):
    r = requests.get(url, stream=True)
    with open(file_path, 'wb') as f:
        total_length = int(r.headers.get('content-length'))
        for chunk in progress.bar(r.iter_content(chunk_size=1024), expected_size=(total_length/1024) + 1):
            if chunk:
                f.write(chunk)
                f.flush()

class Atlas:
    t1 = None #T1 nifti
    t2 = None #T2 nifti
    mask = None #mask nifti
    whitematter_mask = None

    class MNI:
        mni_url = "http://www.bic.mni.mcgill.ca/~vfonov/icbm/2009/old/mni_icbm152_nlin_asym_09c_nifti.zip"

        @classmethod
        def download(self, target_dir, url = mni_url):
            print('Downloading mni_icbm152_nlin_asym_09c_nifti.zip to ', target_dir, ' from ', url)
            download_file(url, os.path.join(target_dir, 'mni_icbm152_nlin_asym_09c_nifti.zip'))
            return self.load(target_dir, download_if_not_found=False)

        @classmethod
        def load(self, target_dir, download_if_not_found=True):
            t1_path = None
            t2_path = None
            mask_path = None

            # Try to find T1 file
            if os.path.isfile(os.path.join(target_dir, 'mni_icbm152_t1_tal_nlin_asym_09c.nii')):
                t1_path = os.path.join(target_dir, 'mni_icbm152_t1_tal_nlin_asym_09c.nii')
            elif os.path.isfile(os.path.join(target_dir, 'mni_icbm152_nlin_asym_09c_nifti.zip')):
                t1_path = os.path.join(target_dir, 'mni_icbm152_t1_tal_nlin_asym_09c.nii')
                with ZipFile(os.path.join(target_dir, 'mni_icbm152_nlin_asym_09c_nifti.zip')) as z:
                    for info in z.infolist():
                        if info.filename == 'mni_icbm152_nlin_asym_09c/mni_icbm152_t1_tal_nlin_asym_09c.nii':
                            info.filename = 'mni_icbm152_t1_tal_nlin_asym_09c.nii'
                            z.extract(info, target_dir)

            # Try to find T2 file
            if os.path.isfile(os.path.join(target_dir, 'mni_icbm152_t2_tal_nlin_asym_09c.nii')):
                t2_path = os.path.join(target_dir, 'mni_icbm152_t2_tal_nlin_asym_09c.nii')
            elif os.path.isfile(os.path.join(target_dir, 'mni_icbm152_nlin_asym_09c_nifti.zip')):
                t2_path = os.path.join(target_dir, 'mni_icbm152_t2_tal_nlin_asym_09c.nii')
                with ZipFile(os.path.join(target_dir, 'mni_icbm152_nlin_asym_09c_nifti.zip')) as z:
                    for info in z.infolist():
                        if info.filename == 'mni_icbm152_nlin_asym_09c/mni_icbm152_t2_tal_nlin_asym_09c.nii':
                            info.filename = 'mni_icbm152_t2_tal_nlin_asym_09c.nii'
                            z.extract(info, target_dir)

            # Try to find Mask file
            if os.path.isfile(os.path.join(target_dir, 'mni_icbm152_t1_tal_nlin_asym_09c_mask.nii')):
                mask_path = os.path.join(target_dir, 'mni_icbm152_t1_tal_nlin_asym_09c_mask.nii')
            elif os.path.isfile(os.path.join(target_dir, 'mni_icbm152_nlin_asym_09c_nifti.zip')):
                mask_path = os.path.join(target_dir, 'mni_icbm152_t1_tal_nlin_asym_09c_mask.nii')
                with ZipFile(os.path.join(target_dir, 'mni_icbm152_nlin_asym_09c_nifti.zip')) as z:
                    for info in z.infolist():
                        if info.filename == 'mni_icbm152_nlin_asym_09c/mni_icbm152_t1_tal_nlin_asym_09c_mask.nii':
                            info.filename = 'mni_icbm152_t1_tal_nlin_asym_09c_mask.nii'
                            z.extract(info, target_dir)

            # Try to find whitematter mask file
            whitematter_path = None
            if os.path.isfile(os.path.join(target_dir, 'mni_icbm152_wm_tal_nlin_asym_09c.nii')):
                whitematter_path = os.path.join(target_dir, 'mni_icbm152_wm_tal_nlin_asym_09c.nii')
            elif os.path.isfile(os.path.join(target_dir, 'mni_icbm152_wm_tal_nlin_asym_09c.zip')):
                mask_path = os.path.join(target_dir, 'mni_icbm152_wm_tal_nlin_asym_09c.nii')
                with ZipFile(os.path.join(target_dir, 'mni_icbm152_wm_tal_nlin_asym_09c.zip')) as z:
                    for info in z.infolist():
                        if info.filename == 'mni_icbm152_wm_tal_nlin_asym_09c/mni_icbm152_wm_tal_nlin_asym_09c.nii':
                            info.filename = 'mni_icbm152_wm_tal_nlin_asym_09c.nii'
                            z.extract(info, target_dir)
            
            # Take a crack at downloading if we can't find a file
            if (t1_path == None or t2_path == None or mask_path == None or whitematter_path == None) and download_if_not_found:
                return self.download(target_dir)

            #Popualte atlas
            atlas = Atlas()
            atlas.t1 = nib.load(t1_path)
            atlas.t2 = nib.load(t2_path)
            atlas.mask = nib.load(mask_path)
            if whitematter_path != None: # Whitematter path is optional at the moment
                atlas.whitematter_mask = nib.load(whitematter_path)

            return atlas



import os
from datetime import date, timedelta, datetime
import nibabel as nib
import numpy as np
import imageio
from PIL import Image
from joblib import Parallel, delayed
import multiprocessing
from pattools.image import m_mode

 #############################
######## INTERPOLATORS ########
 #############################

class _AbstractInterpolator:

    @staticmethod
    def _date_from_path(path):
        datestring = os.path.basename(os.path.dirname(path))
        return date(int(datestring[0:4]), int(datestring[4:6]), int(datestring[6:]))

    @staticmethod
    def interpolated_dates(dates, delta_days):
        ''' Returns a set of interpolated dates from a given list of dates and delta in days.'''
        result_dates = []
        for date1, date2 in zip(dates[:-1], dates[1:]):
            window = date2 - date1
            steps = int(window.days / delta_days)
            for i in range(0, steps):
                result_dates.append(date1+timedelta(days=int(i*delta_days)))
        # Add the last date
        result_dates.append(dates[-1])
        return result_dates


    def data_for_date(self, date, study_dates, study_paths, mask_data):
        ''' This method will return the interpolated data for a given date.
            We are assuming that the study_dates and study_paths are in the same order, which is sorted by date '''
        raise Exception("This is the base interpolator class. Use an implementation")
        

    def interpolated_data_from_dates(self, image_paths, mask_path, dates):
        '''This method will interpolate data for every data given in dates.
        The method will assume the image paths are formatted within a Timeline (so the date
        for each image can be extracted via the _AbstractInterpolator._date_from_path method.'''
        study_dates = [_AbstractInterpolator._date_from_path(p) for p in image_paths]
        mask_data = None
        if mask_path != None and os.path.exists(mask_path):
            mask_data = nib.load(mask_path).get_fdata()
        else:
            print('mask path ' + str(mask_path) + ' does not exist, using no mask')
            mask_data = 1

        for date in dates:
            d, d2 = self.data_for_date(date, study_dates, image_paths, mask_data)
            yield d, d2

    def interpolated_data_from_delta(self, image_paths, mask_path, delta_days):
        ''' This method will interpolate data based on images with a time between returned images
        that is delta_days days'''
        ''' Returns a list of numpy volumes interpolated based on the delta days. All real scans are included.'''
        all_dates = _AbstractInterpolator.interpolated_dates([_AbstractInterpolator._date_from_path(p) for p in image_paths], delta_days)
        return self.interpolated_data_from_dates(image_paths, mask_path, all_dates)



class LinearInterpolator(_AbstractInterpolator):
    '''Interpolates data linearly'''
    def __init__(self):
        super().__init__()

    def interpolate(self, data1, data2, ratio):
        return data1 * (1-ratio) + data2 * ratio
    
    def data_for_date(self, date, study_dates, study_paths, mask_data):
        if date in study_dates:
            data = nib.load(study_paths[study_dates.index(date)]).get_fdata() * mask_data
            return (date, data)
        else:
            # Get the closest two study dates (before and after)
            np_study_dates = np.unique(np.asarray(study_dates))
            np_study_dates = np.argsort(np.abs(np_study_dates - date))
            idx_before = None
            idx_after = None
            for idx in np_study_dates:
                if idx_before == None and study_dates[idx] < date:
                    idx_before = idx
                elif idx_after == None and study_dates[idx] > date:
                    idx_after = idx
            
            # If the date doesn't fall between two dates it's going to be meaningless
            if (idx_before == None or idx_after == None):
                # So we'll just return a zero matrix based on the first entry
                return (date, nib.load(study_paths[0]).get_fdata() * 0)

            # Load the data from those dates
            before_data = nib.load(study_paths[idx_before]).get_fdata() * mask_data
            after_data = nib.load(study_paths[idx_after]).get_fdata() * mask_data
            # Work out the ratio for interpolation
            before_date = study_dates[idx_before]
            after_date = study_dates[idx_after]
            after_delta = after_date - before_date
            delta = date - before_date
            ratio = 0.
            if after_delta != 0:
                ratio = delta / after_delta
            # yield the interpolated result
            return (date, self.interpolate(before_data, after_data, ratio))



class NearestNeighbourInterpolator(LinearInterpolator):
    '''Returns the nearest real scan'''
    def __init__(self):
        super().__init__()

    def interpolate(self, data1, data2, ratio):
        if ratio >= 0.5: return data2
        return data1

class NullInterpolator(LinearInterpolator):
    '''The null iterpolator returns only the masked data (with no interpolation)'''
    def __init__(self):
        super().__init__()

    def interpolate(self, data1, data2, ratio):
        pass

    def interpolated_data(self, image_paths, mask_path, delta_days=None):
        for path in image_paths:
            # Open the nifti file
            img = nib.load(path)
            # Get the data
            data = img.get_fdata()

            if (mask_path != None):
                mask_img = nib.load(mask_path)
                mask_data = mask_img.get_fdata()
                data *= mask_data

            date = _AbstractInterpolator._date_from_path(path)
            yield date, data

 #########################
######## Renderers ########
 #########################

class Renderer:
    '''Renders interpolated data to image files.'''
    interpolator = None
    days_delta = None
    timeline = None

    def __init__(self, interpolator=LinearInterpolator(), days_delta=28):
        self.interpolator = interpolator
        self.days_delta = days_delta

    @staticmethod
    def _get_files(timeline, filter):
        files = []
        for studydate in timeline.datamap:
            files.extend([
                os.path.join(timeline.path, studydate, fm.processed_file)
                for fm in timeline.datamap[studydate]
                if fm.filter_name == filter.name])
        return files

    def render(self, timeline, path, overwrite=True, output_type='gray16'):
        '''Write images to path given based on a timeline. Files will be interpolated and rendered to <path>/<filter>/<cor|sag|ax>/<date>/'''
        print('timeline.path', timeline.path)
        print('timeline.whitematter_mask', timeline.whitematter_mask)
        mask_path = None
        whitematter_mask_path = None
        if timeline.path != None and timeline.brain_mask != None:
             mask_path = os.path.join(timeline.path, timeline.brain_mask)
        if timeline.whitematter_mask != None:
            whitematter_mask_path = os.path.join(timeline.path, timeline.whitematter_mask)

        for filter in timeline.filters:
            files = Renderer._get_files(timeline, filter)
            self.render_all(
                files,
                mask_path, 
                os.path.join(path, filter.name), 
                timeline.study_dates(), 
                whitematter_mask_path=whitematter_mask_path,
                overwrite=overwrite,
                output_type=output_type)

    def render_new_studies(self, timeline, path, output_type='gray16'):
        '''Write images to path given based on a timeline. Files will be interpolated and rendered to <path>/<filter>/<cor|sag|ax>/<date>/'''
        self.render(timeline, path, overwrite=False, output_type=output_type)

    @staticmethod
    def write_images(data, folder, slice_type, min_val, max_val, image_mode='gray16-rgbstuffed'):
        print('writing images...')
        if not os.path.exists(folder):
            os.makedirs(folder, mode=0o777)

        data_cp = np.flip(data, 0)
        count = 0
        if slice_type == 'sag':
            count = data.shape[0]
            for i in range(data.shape[0]):
                Renderer.write_image(data_cp[i,:,:], os.path.join(folder, f'{i}.png'), min_val, max_val, image_mode)

        elif slice_type == 'cor':
            count = data.shape[1]
            for j in range(data.shape[1]):
                Renderer.write_image(data_cp[:,j,:], os.path.join(folder, f'{j}.png'), min_val, max_val, image_mode)

        elif slice_type == 'ax':
            count = data.shape[2]
            for k in range(data.shape[2]):
                Renderer.write_image(data_cp[:,:,k], os.path.join(folder, f'{k}.png'), min_val, max_val, image_mode)

        return count

    @staticmethod
    def write_image(slice, location, min, max, mode='gray8'):
        '''Write a single slice to an image at the given location'''
        if mode == 'gray16':
            Renderer.write_grayscale16_image(slice, location, min, max)
        elif mode == 'gray8':
            Renderer.write_grayscale8_image(slice, location, min, max)
        elif mode == 'gray16-rgbstuffed':
            Renderer.write_grayscale16_rgbstuffed_image(slice, location, min, max)

    @staticmethod
    def write_grayscale16_image(slice, location, min, max):
        '''Write a single slice to an image at the given location as 16bit greyscale'''
        # For non background slices we want to normalize the range to fit in uint16
        output = np.flip(slice.T).copy()
        output = output.astype(np.uint16)
        Image.fromarray(output).save(location)

    @staticmethod
    def write_grayscale16_rgbstuffed_image(slice, location, min, max):
        '''This will preserve the 16 bits of data to allow for windowing, but will output a standard RGB image.
        This method allows using standard libraries that expect RGB without losing data, but will need to be
        processed and windowed before output'''
        # For non background slices we want to normalize the range to fit in uint16
        output = np.flip(slice.T).copy()
        output = output.astype(np.uint16)

        rgbout = np.zeros((output.shape[0], output.shape[1], 3), np.uint8)
        rgbout[:,:,1] = (output >> 8).astype(np.uint8)
        
        rgbout[:,:,2] = (output - (rgbout[:,:,1] << 8)).astype(np.uint8)
        Image.fromarray(rgbout).save(location)

    @staticmethod
    def write_grayscale8_image(slice, location, min, max):
        '''Write a single slice to an image at the given location as 16bit greyscale'''
        # For non background slices we want to normalize the range to fit in uint16
        output = np.flip(slice.T).copy()
        output = output.astype(np.uint8)
        Image.fromarray(output).save(location)

    @staticmethod
    def _render_volume(date, volume, path, overwrite=True, whitematter_mask=1, output_type='gray16'):
        '''Render every slice in a volume along 3 axis.'''
        # Set the min to be 0
        volume = volume - np.amin(volume)
        # Make sure the curve falls between 0 and 400
        max_val = np.amax(volume)
        #if max_val > 255:
        #    volume = volume / max_val * 255
        #volume = volume.astype(np.uint8)

        # Output paths
        sag_path = os.path.join(path, date.strftime('%Y%m%d'), 'sag')
        cor_path = os.path.join(path, date.strftime('%Y%m%d'), 'cor')
        ax_path = os.path.join(path, date.strftime('%Y%m%d'), 'ax')
        # Write images if the folder doesn't exist or overwrite is true
        if overwrite or os.path.exists(sag_path) == False:
            Renderer.write_images(volume, sag_path, 'sag', 0, max_val)
        if overwrite or os.path.exists(cor_path) == False:
            Renderer.write_images(volume, cor_path, 'cor', 0, max_val)
        if overwrite or os.path.exists(ax_path) == False:
            Renderer.write_images(volume, ax_path, 'ax', 0, max_val)

    def render_all(self, files, mask_path, path, dates, whitematter_mask_path=None, overwrite=True, output_type='gray16'):
        # Load the whitematter mask to use (or use 1, i.e. no mask if none is found)
        whitematter_mask = 1
        print('??whitematter_mask_path', whitematter_mask_path)
        if whitematter_mask_path !=  None:
            whitematter_mask = nib.load(whitematter_mask_path).get_fdata()

        '''Render all volumes using supplied brain mask'''
        Parallel(n_jobs=multiprocessing.cpu_count())(
            delayed(Renderer._render_volume)(date, volume, path, overwrite=overwrite, whitematter_mask=whitematter_mask, output_type=output_type)
            for date, volume in self.interpolator.interpolated_data_from_dates(files, mask_path, dates))

    def render_new(self, files, mask_path, path, dates,  whitematter_mask_path=None, output_type='gray16'):
        self.render_all(files, mask_path, path, dates, whitematter_mask_path=whitematter_mask_path, overwrite=False, output_type=output_type)


#class VisTarsierRenderer(Renderer):
#    def render_all(self, files, mask_path, path, dates, whitematter_mask_path=None, overwrite=True):
#        #dates = _AbstractInterpolator.interpolated_dates(self.timeline.study_dates)
#        '''Render all volumes using supplied brain mask'''
#        Parallel(n_jobs=multiprocessing.cpu_count())(
#            delayed(Renderer._render_volume)(date, volume, path, overwrite=True)
#            for date, volume in self.interpolator.interpolated_data_from_dates(files, mask_path, dates))

class MMode():
    ''' M Mode renders a 2D view of the 4D timeline data between two spatial points (i.e. a line through time)'''
    timeline = None
    _start = None
    _end = None
    _lines = None
    
    def __init__(self, timeline, start, end):
        ''' M Mode renderer requires a timeline, start point and end point'''
        self.timeline = timeline
        timeline._load_datamap()
        self._start = start
        self._end = end

    def _calculate_lines(self, height=100):
        ''' This will give us a bunch of 1D lines, one for each scan in the timeline. '''
        if self._start == None or self._end == None:
            raise ValueError('Start and end points cannot be None')
        
        self._lines = {}
        for studydate in self.timeline.datamap:
            for fm in self.timeline.datamap[studydate]:
                path = os.path.join(self.timeline.path, studydate, fm.processed_file)
                try:
                    img = nib.load(path).get_fdata()
           
                    #resolve negative indexing (a little wasteful to do it inside the loop but we have an image shape and it's cheap)
                    for i in range(3):
                        if self._start[i] < 0:
                            self._start[i] += img.shape[i]
                        if self._end[i] < 0:
                            self._end[i] += img.shape[i]
                    # Calculate line         
                    line = m_mode(img, self._start, self._end, out_size=height)
                    if fm.filter_name not in self._lines:
                        self._lines[fm.filter_name] = {}
                    self._lines[fm.filter_name][studydate] = line
                except Exception as e:
                    print(f'Something went wrong with {fm.filter_name}/{studydate}')
                    print(e)
        return self._lines

    def _parse_date(self, date_time_str):
        ''' Converts a timeline date into a timeline for norming. '''
        return datetime.strptime(date_time_str, '%Y%m%d').timestamp()

    # TODO: This function is trying to do two things (output the file and return the dict), they should be separated.
    def render(self, filter_name, outfile=None, line_width=5, width=400, height=100):
        ''' This function will render a 2D view of a given filter to the outfile.
            Background will be black in an image defined by height and width.
            Return value is a list of dictionaries with normed dates and line values. '''
        # Calculate the lines as needed
        if self._lines == None or len(self._lines.keys()) == 0 or self._lines[next(self._lines.keys())].shape[1] != height:
            self._calculate_lines(height)
                
        # Check input
        if self._lines == None or self._start == None or self._end == None:
            raise ValueError('Points have not been set correctly')
        if filter_name not in self._lines or len(self._lines[filter_name].keys()) == 0:
            raise ValueError("Filter doesn't seem to exist")
        
        lines = self._lines[filter_name]
        
        outdata = np.zeros((width, height))
        dates = []
        
        # Parse the date values
        max_date = float('-inf')
        min_date = float('inf')
        for date_string in lines.keys():
            d = self._parse_date(date_string)
            dates.append({'date_string':date_string, 'date':d, 'line':lines[date_string]})
            min_date = min(min_date, d)
            max_date = max(max_date, d)
        
        # Write the line to the output image
        max_date -= min_date
        for ddict in dates:
            normed_date = (ddict['date'] - min_date) / max_date
            index = int(normed_date * (width-line_width))
            for i in range(line_width):
                if index+i < outdata.shape[0] and index+1 >= 0:
                    outdata[index+i,:] = ddict['line']
        
        # Write out file
        if outfile != None:
            # Norm data to 16bit signed int
            outdata -= np.min(outdata)
            maxval = np.max(outdata)
            if maxval != 0:
                outdata /= maxval
                outdata *= 255
            
            outimg = Image.fromarray(outdata.T.astype(np.uint8), mode='L')#, mode='I;16')
            outimg.save(outfile)
            print(np.array(outimg))

        return dates



import os
import subprocess

FSLPATH = ''
if 'FSLPATH' not in os.environ:
    print('***************** OH NO! ********************')
    print("Can't find FSLPATH environment value. D:")
    print("pattools.fsl is just a thin wrapper around a couple of FSL tools that you need to install yourself.")
    print("We'll assume it's here somewhere (and in your PATH), but if I don't work try the docker image.")
    print('***************** /OH NO! *******************')
else:
    FSLPATH = os.environ['FSLPATH']

def bet(input, output):
    betpath = os.path.join(FSLPATH, 'fsl5.0-bet')
    return subprocess.Popen([betpath, input, output, '-f','0.4','-R'])


import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="pattools",
    version="0.1.3",
    author="plwp",
    author_email="pat@plwp.net",
    description="Toolkit for neuro-imaging data manipulation and automation",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/plwp/pat_tools",
    packages=['pattools'],
    install_requires=[
          'pydicom',
          'pynetdicom',
          'dicom2nifti',
          'requests',
          'nibabel',
          'clint',
          'imageio',
          'joblib',
          'h5py',
          'numpy',
          'scikit-learn',
          'scikit-image'
      ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
)



from tempfile import TemporaryDirectory
from pattools.pacs import Patient
from pattools._timelinefilter import flair_filter
from pattools.ants import n4_bias_correct, affine_registration
from pattools.fsl import bet
import os
import subprocess
import numpy as np

def latest_flair_pair(patient_id, scp_settings):
    ''' Get the 2 most recent FLAIR studies for a patient. Returns None if we can't find 2 studies'''
    p = Patient(patient_id, scp_settings=scp_settings)
    studies = p.find_studies()
    studies.sort(key=lambda s: s.study_date, reverse=True)
    fil = flair_filter()
    latest_flairs = []
    for s in studies:
        flair = fil.filter(s)
        if flair != None:
            latest_flairs.append(flair)
        if len(latest_flairs) >= 2:
            break
    if len(latest_flairs) == 2:
        return latest_flairs
    else:
        return None

def preprocess_pair(floating, fixed, tmp_dir):
    """Pre processes the nifti images using:
    ANTs N4BiasFieldCorrection -> FSL bet -> ANTs antsRegistration
    Parameters
    ----------
    floating : string
        Path to the floating nifti image
    fixed : string
        Path to the fixed nifti image
    Returns
    -------
    (floating, fixed) : tuple
        floating : string
            Path to the pre-processed floating image
        fixed : string
            Path to the pre-processed floating image
    """
    with TemporaryDirectory() as local_temp:
        # ANTS N4 Bias correction
        outfloat = os.path.join(local_temp, 'n4float.nii.gz')
        outfixed = os.path.join(local_temp, 'n4fixed.nii.gz')
        p1 = n4_bias_correct(floating, outfloat)
        p2 = n4_bias_correct(fixed, outfixed)
        p1.wait()
        p2.wait()
        # FSL BET
        floating = outfloat
        fixed = outfixed
        outfloat = os.path.join(local_temp, 'betfloat.nii.gz')
        outfixed = os.path.join(tmp_dir, 'fixed-processed.nii.gz') # Registration won't change this
        p1 = bet(floating, outfloat)
        p2 = bet(fixed, outfixed)
        p1.wait()
        p2.wait()
        # Coregistration
        floating = outfloat
        fixed = outfixed
        outfloat = os.path.join(tmp_dir, 'floating-processed.nii.gz')
        affine_registration(floating, fixed, outfloat).wait()
        
        return outfloat, outfixed

def vistarsier_compare(c, p, min_val=-1., max_val=5., min_change=0.8, max_change=3.):
    """ VisTarsier's compare operation
    Parameters
    ----------
    c : ndarray
        The current volume
    p : ndarray
        The prior volume
    min_val : float
        The minimum value (measured in standard deviations) to consider
    max_val : float
        The maximum value (measured in standard deviations) to consider
    min_change : float
        The minimum change of value (measured in standard deviations) to consider
    max_change : float
        The maximum change of value (measured in standard deviations) to consider
    Returns
    -------
    change : ndarray
        The relevant change in signal.
    """
    # Get standard deviations for current and prior
    pstd = p.std()
    cstd = c.std()
    # Align prior standard deviation to current
    p = ((p - p.mean()) / pstd) * cstd + c.mean()

    #Calculate change
    change = c - p
    # Ignore change outside of minimuim and maximum values
    change[c < min_val*cstd] = 0
    change[p < min_val*cstd] = 0
    change[c > max_val*cstd] = 0
    change[p > max_val*cstd] = 0
    change[np.abs(change) < min_change*cstd] = 0
    change[np.abs(change) > max_change*cstd] = 0
    return change